interface Schema {
    [K: string]: Schema | string;
}
type InferLiteralString<T, TD extends string> = T extends infer K ? K extends string ? K extends '' ? `${TD}` : `${K}` : `${TD}` : `${TD}`;
type ScopePrefixer<T, S> = T extends infer K ? K extends string ? K extends '' ? '' : `${K}${InferLiteralString<S, '.'>}` : never : never;
type UnionScopes<T, Prefix extends string = '', Spacer extends string = '.'> = T extends Record<string, any> ? {
    [K in keyof T]: K extends string ? T[K] extends Record<string, any> ? UnionScopes<T[K], `${ScopePrefixer<Prefix, Spacer>}${K}`, InferLiteralString<Spacer, '.'>> : `${ScopePrefixer<Prefix, Spacer>}${K}` : never;
}[keyof T] : Prefix;
type PermitNode = {
    scope: string;
    granted: boolean;
    desc: string;
    toggle: () => void;
};
type TransformSchema<T, TValue> = T extends Schema ? {
    [P in keyof T]: T[P] extends Schema ? TransformSchema<T[P], TValue> : TValue;
} : TValue;
type SchemaOptions = {
    /** add prefix to scope. default to '' (none) */
    prefix: string;
    /** string to use for spacer for scope. default to '.' */
    spacer: string;
};

type SchemaNode = {
    scope: string;
    description: string;
};
declare class AccessSchema<T extends Schema = Schema> {
    private _options;
    private _schema;
    private _scopes;
    constructor(schema: T, options?: Partial<SchemaOptions>);
    private createSchemaNode;
    private resetSchemaNode;
    private get scopesList();
    /** utility function for finding scope */
    private _get;
    /** readonly schema options */
    get options(): SchemaOptions;
    /** readonly AccessNodes  */
    get nodes(): TransformSchema<T, string>;
    /** export schema to list of scope and description */
    export(): SchemaNode[];
    /** get schema node by string scope */
    get(scope: string): SchemaNode | undefined;
    /** utility function for checking if there is a schema node that have an exact match with given scope */
    has(scope: string): boolean;
    /** find schema node that has the given substring in scope or description. */
    search(substring: string, fn?: (n: SchemaNode) => boolean): SchemaNode[];
    /** find all child of the given string scope */
    getChild(scope: string): SchemaNode[];
    /** Map over schema and transform leaf node by given handler */
    map<V>(handler: (node: SchemaNode) => V): TransformSchema<T, V>;
    /**
     * Create permit from schema with given granted scopes
     */
    permit(scopes: string[] | SchemaNode[]): {
        get: <Prefix extends string = "", Spacer extends string = "">(scope: UnionScopes<T, Prefix, Spacer>) => PermitNode | undefined;
        has: <Prefix_1 extends string = "", Spacer_1 extends string = "">(scope: UnionScopes<T, Prefix_1, Spacer_1>) => boolean;
        search: (substring: string, handle: (node: PermitNode) => void) => void;
        getChild: (scope: string) => PermitNode[];
        reset: () => void;
        grant: (...scopes: string[] | SchemaNode[]) => void;
        deny: (...scopes: string[] | SchemaNode[]) => void;
        check(scope: string): {
            readonly exact: boolean;
            readonly hasChild: boolean;
            readonly hasAnyChildGranted: boolean;
            readonly hasAllChildGranted: boolean;
            readonly hasAllChildDenied: boolean;
        };
        readonly export: {
            readonly all: {
                scope: string;
                desc: string;
            }[];
            readonly granted: {
                scope: string;
                desc: string;
            }[];
            readonly denied: {
                scope: string;
                desc: string;
            }[];
        };
        readonly granted: {
            scope: string;
            desc: string;
        }[];
        readonly denied: {
            scope: string;
            desc: string;
        }[];
        readonly nodes: TransformSchema<T, PermitNode>;
        readonly can: TransformSchema<T, boolean>;
    };
}

type PermitScope = Omit<PermitNode, 'toggle' | 'granted'>;
type Permit<T extends Schema> = ReturnType<typeof permit$1<T>>;
declare function permit$1<T extends Schema>(access: AccessSchema<T>, values: string[]): {
    get: <Prefix extends string = "", Spacer extends string = "">(scope: UnionScopes<T, Prefix, Spacer>) => PermitNode | undefined;
    has: <Prefix_1 extends string = "", Spacer_1 extends string = "">(scope: UnionScopes<T, Prefix_1, Spacer_1>) => boolean;
    search: (substring: string, handle: (node: PermitNode) => void) => void;
    getChild: (scope: string) => PermitNode[];
    reset: () => void;
    grant: (...scopes: string[] | SchemaNode[]) => void;
    deny: (...scopes: string[] | SchemaNode[]) => void;
    /** utility for checking string scope permit value */
    check(scope: string): {
        readonly exact: boolean;
        readonly hasChild: boolean;
        readonly hasAnyChildGranted: boolean;
        readonly hasAllChildGranted: boolean;
        readonly hasAllChildDenied: boolean;
    };
    /** getter for exporting scopes */
    readonly export: {
        readonly all: PermitScope[];
        readonly granted: PermitScope[];
        readonly denied: PermitScope[];
    };
    /** readonly enabled access node */
    readonly granted: PermitScope[];
    /** readonly disabled access node */
    readonly denied: PermitScope[];
    /** readonly AccessNodes  */
    readonly nodes: TransformSchema<T, PermitNode>;
    /** readonly assertion utility for granted scopes  */
    readonly can: TransformSchema<T, boolean>;
};

/**
 * Create access control schema
 *
 * @param {T} schema
 * @param {Partial<SchemaOptions>} [options={}]
 * @return {AccessSchema} access schema
 */
declare function createSchema<T extends Schema>(schema: T, options?: Partial<SchemaOptions>): AccessSchema<T>;
type PermitOptions = {
    granted: string[] | SchemaNode[];
    denied: string[] | SchemaNode[];
} & SchemaOptions;
/**
 * Create user permit
 *
 * @export
 * @template T
 * @param {AccessSchema<T>} schema
 * @param {(string[] | SchemaNode[])} scopes
 * @return {Permit}
 */
declare function createPermit<T extends Schema>(schema: T, options?: Partial<PermitOptions>): {
    get: <Prefix extends string = "", Spacer extends string = "">(scope: UnionScopes<T, Prefix, Spacer>) => PermitNode | undefined;
    has: <Prefix_1 extends string = "", Spacer_1 extends string = "">(scope: UnionScopes<T, Prefix_1, Spacer_1>) => boolean;
    search: (substring: string, handle: (node: PermitNode) => void) => void;
    getChild: (scope: string) => PermitNode[];
    reset: () => void;
    grant: (...scopes: string[] | SchemaNode[]) => void;
    deny: (...scopes: string[] | SchemaNode[]) => void;
    check(scope: string): {
        readonly exact: boolean;
        readonly hasChild: boolean;
        readonly hasAnyChildGranted: boolean;
        readonly hasAllChildGranted: boolean;
        readonly hasAllChildDenied: boolean;
    };
    readonly export: {
        readonly all: {
            scope: string;
            desc: string;
        }[];
        readonly granted: {
            scope: string;
            desc: string;
        }[];
        readonly denied: {
            scope: string;
            desc: string;
        }[];
    };
    readonly granted: {
        scope: string;
        desc: string;
    }[];
    readonly denied: {
        scope: string;
        desc: string;
    }[];
    readonly nodes: TransformSchema<T, PermitNode>;
    readonly can: TransformSchema<T, boolean>;
};
/**
 * Create guard utility that check based on given guard function
 *
 * @export
 * @template T
 * @template P
 * @param {AccessSchema<T>} schema
 * @param {(permit: Permit<T, P>) => boolean} guard
 * @return {*}
 */
declare function createGuard<T extends Schema>(schema: AccessSchema<T>, guard: (permit: Permit<T>) => boolean): {
    grant(scopes: string[]): void;
    deny(scopes: string[]): void;
    readonly isAllowed: boolean;
    readonly isDenied: boolean;
};

type ZodType<T> = {
    parse: (data: any) => T;
    array: () => ZodType<T[]>;
};
type parseOptions<T> = {
    default?: T;
    parseHandler?: (value: DataType) => T;
};
type DataRecord = {
    [P in string]: DataType;
};
type DataType = DataRecord | DataRecord[] | any[] | string | number | boolean | null | undefined;
type Handler<T> = (row: Inferer) => T;
type Inferer = ReturnType<typeof infer>;
declare function infer(data: DataType): {
    map: <T>(path: string | undefined, handler: Handler<T>) => T[];
    array: <T_1 = DataType>(path: string, schema: ZodType<T_1> | Handler<T_1>, options?: parseOptions<T_1[]>) => T_1[];
    object: <T_2>(path: string, schema: ZodType<T_2> | Handler<T_2>, options?: parseOptions<T_2>) => T_2;
    string: (path?: string, options?: parseOptions<string>) => string;
    number: (path?: string, options?: parseOptions<number> & {
        float?: boolean;
    }) => number;
    boolean: (path?: string, options?: parseOptions<boolean>) => boolean;
    nullable: (path?: string) => {
        string: (options?: parseOptions<string>) => string | null;
        number: (options?: parseOptions<number> & {
            float?: boolean;
        }) => number | null;
        boolean: (options?: parseOptions<boolean>) => boolean | null;
        stringOrNumber: (options?: parseOptions<string | number> & {
            float?: boolean;
        }) => string | number | null;
        array: <T_3 = DataType>(handler: ZodType<T_3> | Handler<T_3>, options?: parseOptions<T_3[]>) => T_3[] | null;
        object: <T_4>(handler: ZodType<T_4> | Handler<T_4>, options?: parseOptions<T_4>) => T_4 | null;
        map: <T_5>(handler: Handler<T_5>) => T_5[] | null;
    };
    get: (path?: string) => DataType;
    stringOrNumber: (path?: string, options?: parseOptions<string | number> & {
        float?: boolean;
    }) => string | number;
    readonly raw: DataType;
    has(path: string): boolean;
};

type ContextRoot = {
    scopes?: string[];
    protected?: boolean;
};
type Context<T> = T & ContextRoot;
type ContextValue<T, V> = T & {
    value: V;
} & ContextRoot;
declare function createProcedure<T extends Record<string, any>, I extends any[], V = undefined>(ctx: () => Context<T>, fn?: (context: Context<T>, ...args: I) => V): {
    fn: (...input: I) => V;
    pipe<Value>(func: (context: ContextValue<T, V>, ...input: I) => Value): {
        readonly fn: (...input: I) => Value;
        pipe<Revalue>(fn: (context: ContextValue<T, Value>, ...input: I) => Revalue): {
            readonly fn: (...input: I) => Revalue;
            pipe<Revalue_1>(fn: (context: ContextValue<T, Revalue>, ...input: I) => Revalue_1): {
                readonly fn: (...input: I) => Revalue_1;
                pipe<Revalue_2>(fn: (context: ContextValue<T, Revalue_1>, ...input: I) => Revalue_2): {
                    readonly fn: (...input: I) => Revalue_2;
                    pipe<Revalue_3>(fn: (context: ContextValue<T, Revalue_2>, ...input: I) => Revalue_3): {
                        readonly fn: (...input: I) => Revalue_3;
                        pipe<Revalue_4>(fn: (context: ContextValue<T, Revalue_3>, ...input: I) => Revalue_4): {
                            readonly fn: (...input: I) => Revalue_4;
                            pipe<Revalue_5>(fn: (context: ContextValue<T, Revalue_4>, ...input: I) => Revalue_5): {
                                readonly fn: (...input: I) => Revalue_5;
                                pipe<Revalue_6>(fn: (context: ContextValue<T, Revalue_5>, ...input: I) => Revalue_6): {
                                    readonly fn: (...input: I) => Revalue_6;
                                    pipe<Revalue_7>(fn: (context: ContextValue<T, Revalue_6>, ...input: I) => Revalue_7): {
                                        readonly fn: (...input: I) => Revalue_7;
                                        pipe<Revalue_8>(fn: (context: ContextValue<T, Revalue_7>, ...input: I) => Revalue_8): {
                                            readonly fn: (...input: I) => Revalue_8;
                                            pipe<Revalue_9>(fn: (context: ContextValue<T, Revalue_8>, ...input: I) => Revalue_9): {
                                                readonly fn: (...input: I) => Revalue_9;
                                                pipe<Revalue_10>(fn: (context: ContextValue<T, Revalue_9>, ...input: I) => Revalue_10): any;
                                            };
                                        };
                                    };
                                };
                            };
                        };
                    };
                };
            };
        };
    };
    protect<S extends Schema>(schema: AccessSchema<S>, guard: (permit: {
        get: <Prefix extends string = "", Spacer extends string = "">(scope: UnionScopes<S, Prefix, Spacer>) => PermitNode | undefined;
        has: <Prefix_1 extends string = "", Spacer_1 extends string = "">(scope: UnionScopes<S, Prefix_1, Spacer_1>) => boolean;
        search: (substring: string, handle: (node: PermitNode) => void) => void;
        getChild: (scope: string) => PermitNode[];
        reset: () => void;
        grant: (...scopes: string[] | SchemaNode[]) => void;
        deny: (...scopes: string[] | SchemaNode[]) => void;
        check(scope: string): {
            readonly exact: boolean;
            readonly hasChild: boolean;
            readonly hasAnyChildGranted: boolean;
            readonly hasAllChildGranted: boolean;
            readonly hasAllChildDenied: boolean;
        };
        readonly export: {
            readonly all: {
                scope: string;
                desc: string;
            }[];
            readonly granted: {
                scope: string;
                desc: string;
            }[];
            readonly denied: {
                scope: string;
                desc: string;
            }[];
        };
        readonly granted: {
            scope: string;
            desc: string;
        }[];
        readonly denied: {
            scope: string;
            desc: string;
        }[];
        readonly nodes: TransformSchema<S, PermitNode>;
        readonly can: TransformSchema<S, boolean>;
    }) => boolean): {
        fn: (...input: I) => V;
        pipe<Value>(func: (context: ContextValue<T, V>, ...input: I) => Value): {
            readonly fn: (...input: I) => Value;
            pipe<Revalue>(fn: (context: ContextValue<T, Value>, ...input: I) => Revalue): {
                readonly fn: (...input: I) => Revalue;
                pipe<Revalue_1>(fn: (context: ContextValue<T, Revalue>, ...input: I) => Revalue_1): {
                    readonly fn: (...input: I) => Revalue_1;
                    pipe<Revalue_2>(fn: (context: ContextValue<T, Revalue_1>, ...input: I) => Revalue_2): {
                        readonly fn: (...input: I) => Revalue_2;
                        pipe<Revalue_3>(fn: (context: ContextValue<T, Revalue_2>, ...input: I) => Revalue_3): {
                            readonly fn: (...input: I) => Revalue_3;
                            pipe<Revalue_4>(fn: (context: ContextValue<T, Revalue_3>, ...input: I) => Revalue_4): {
                                readonly fn: (...input: I) => Revalue_4;
                                pipe<Revalue_5>(fn: (context: ContextValue<T, Revalue_4>, ...input: I) => Revalue_5): {
                                    readonly fn: (...input: I) => Revalue_5;
                                    pipe<Revalue_6>(fn: (context: ContextValue<T, Revalue_5>, ...input: I) => Revalue_6): {
                                        readonly fn: (...input: I) => Revalue_6;
                                        pipe<Revalue_7>(fn: (context: ContextValue<T, Revalue_6>, ...input: I) => Revalue_7): {
                                            readonly fn: (...input: I) => Revalue_7;
                                            pipe<Revalue_8>(fn: (context: ContextValue<T, Revalue_7>, ...input: I) => Revalue_8): {
                                                readonly fn: (...input: I) => Revalue_8;
                                                pipe<Revalue_9>(fn: (context: ContextValue<T, Revalue_8>, ...input: I) => Revalue_9): {
                                                    readonly fn: (...input: I) => Revalue_9;
                                                    pipe<Revalue_10>(fn: (context: ContextValue<T, Revalue_9>, ...input: I) => Revalue_10): any;
                                                };
                                            };
                                        };
                                    };
                                };
                            };
                        };
                    };
                };
            };
        };
        permit: (scopes: string[]) => {
            fn: (...input: I) => V;
            pipe<Value>(func: (context: ContextValue<T, V>, ...input: I) => Value): {
                readonly fn: (...input: I) => Value;
                pipe<Revalue>(fn: (context: ContextValue<T, Value>, ...input: I) => Revalue): {
                    readonly fn: (...input: I) => Revalue;
                    pipe<Revalue_1>(fn: (context: ContextValue<T, Revalue>, ...input: I) => Revalue_1): {
                        readonly fn: (...input: I) => Revalue_1;
                        pipe<Revalue_2>(fn: (context: ContextValue<T, Revalue_1>, ...input: I) => Revalue_2): {
                            readonly fn: (...input: I) => Revalue_2;
                            pipe<Revalue_3>(fn: (context: ContextValue<T, Revalue_2>, ...input: I) => Revalue_3): {
                                readonly fn: (...input: I) => Revalue_3;
                                pipe<Revalue_4>(fn: (context: ContextValue<T, Revalue_3>, ...input: I) => Revalue_4): {
                                    readonly fn: (...input: I) => Revalue_4;
                                    pipe<Revalue_5>(fn: (context: ContextValue<T, Revalue_4>, ...input: I) => Revalue_5): {
                                        readonly fn: (...input: I) => Revalue_5;
                                        pipe<Revalue_6>(fn: (context: ContextValue<T, Revalue_5>, ...input: I) => Revalue_6): {
                                            readonly fn: (...input: I) => Revalue_6;
                                            pipe<Revalue_7>(fn: (context: ContextValue<T, Revalue_6>, ...input: I) => Revalue_7): {
                                                readonly fn: (...input: I) => Revalue_7;
                                                pipe<Revalue_8>(fn: (context: ContextValue<T, Revalue_7>, ...input: I) => Revalue_8): {
                                                    readonly fn: (...input: I) => Revalue_8;
                                                    pipe<Revalue_9>(fn: (context: ContextValue<T, Revalue_8>, ...input: I) => Revalue_9): {
                                                        readonly fn: (...input: I) => Revalue_9;
                                                        pipe<Revalue_10>(fn: (context: ContextValue<T, Revalue_9>, ...input: I) => Revalue_10): any;
                                                    };
                                                };
                                            };
                                        };
                                    };
                                };
                            };
                        };
                    };
                };
            };
        };
    };
};
declare function createContext<T>(ctx: () => Context<T>, immidiate?: boolean): () => Context<T>;
type InterfaceConfig<TC, SC> = {
    sharedContext: () => Context<SC>;
    context: () => Context<TC>;
    onError?: (error: Error) => void;
    onCall?: (value: any) => void;
};
declare function createInterface<C, V>(config?: Partial<InterfaceConfig<C, V>>): {
    procedure: <Input extends any[], Value = void>(fn?: (context: Context<C & V>, ...args: Input) => Value) => {
        fn: (...input: Input) => Value;
        pipe<Value_1>(func: (context: ContextValue<Context<C & V>, Value>, ...input: Input) => Value_1): {
            readonly fn: (...input: Input) => Value_1;
            pipe<Revalue>(fn: (context: ContextValue<Context<C & V>, Value_1>, ...input: Input) => Revalue): {
                readonly fn: (...input: Input) => Revalue;
                pipe<Revalue_1>(fn: (context: ContextValue<Context<C & V>, Revalue>, ...input: Input) => Revalue_1): {
                    readonly fn: (...input: Input) => Revalue_1;
                    pipe<Revalue_2>(fn: (context: ContextValue<Context<C & V>, Revalue_1>, ...input: Input) => Revalue_2): {
                        readonly fn: (...input: Input) => Revalue_2;
                        pipe<Revalue_3>(fn: (context: ContextValue<Context<C & V>, Revalue_2>, ...input: Input) => Revalue_3): {
                            readonly fn: (...input: Input) => Revalue_3;
                            pipe<Revalue_4>(fn: (context: ContextValue<Context<C & V>, Revalue_3>, ...input: Input) => Revalue_4): {
                                readonly fn: (...input: Input) => Revalue_4;
                                pipe<Revalue_5>(fn: (context: ContextValue<Context<C & V>, Revalue_4>, ...input: Input) => Revalue_5): {
                                    readonly fn: (...input: Input) => Revalue_5;
                                    pipe<Revalue_6>(fn: (context: ContextValue<Context<C & V>, Revalue_5>, ...input: Input) => Revalue_6): {
                                        readonly fn: (...input: Input) => Revalue_6;
                                        pipe<Revalue_7>(fn: (context: ContextValue<Context<C & V>, Revalue_6>, ...input: Input) => Revalue_7): {
                                            readonly fn: (...input: Input) => Revalue_7;
                                            pipe<Revalue_8>(fn: (context: ContextValue<Context<C & V>, Revalue_7>, ...input: Input) => Revalue_8): {
                                                readonly fn: (...input: Input) => Revalue_8;
                                                pipe<Revalue_9>(fn: (context: ContextValue<Context<C & V>, Revalue_8>, ...input: Input) => Revalue_9): {
                                                    readonly fn: (...input: Input) => Revalue_9;
                                                    pipe<Revalue_10>(fn: (context: ContextValue<Context<C & V>, Revalue_9>, ...input: Input) => Revalue_10): any;
                                                };
                                            };
                                        };
                                    };
                                };
                            };
                        };
                    };
                };
            };
        };
        protect<S extends Schema>(schema: AccessSchema<S>, guard: (permit: {
            get: <Prefix extends string = "", Spacer extends string = "">(scope: UnionScopes<S, Prefix, Spacer>) => PermitNode | undefined;
            has: <Prefix_1 extends string = "", Spacer_1 extends string = "">(scope: UnionScopes<S, Prefix_1, Spacer_1>) => boolean;
            search: (substring: string, handle: (node: PermitNode) => void) => void;
            getChild: (scope: string) => PermitNode[];
            reset: () => void;
            grant: (...scopes: string[] | SchemaNode[]) => void;
            deny: (...scopes: string[] | SchemaNode[]) => void;
            check(scope: string): {
                readonly exact: boolean;
                readonly hasChild: boolean;
                readonly hasAnyChildGranted: boolean;
                readonly hasAllChildGranted: boolean;
                readonly hasAllChildDenied: boolean;
            };
            readonly export: {
                readonly all: {
                    scope: string;
                    desc: string;
                }[];
                readonly granted: {
                    scope: string;
                    desc: string;
                }[];
                readonly denied: {
                    scope: string;
                    desc: string;
                }[];
            };
            readonly granted: {
                scope: string;
                desc: string;
            }[];
            readonly denied: {
                scope: string;
                desc: string;
            }[];
            readonly nodes: TransformSchema<S, PermitNode>;
            readonly can: TransformSchema<S, boolean>;
        }) => boolean): {
            fn: (...input: Input) => Value;
            pipe<Value_1>(func: (context: ContextValue<Context<C & V>, Value>, ...input: Input) => Value_1): {
                readonly fn: (...input: Input) => Value_1;
                pipe<Revalue>(fn: (context: ContextValue<Context<C & V>, Value_1>, ...input: Input) => Revalue): {
                    readonly fn: (...input: Input) => Revalue;
                    pipe<Revalue_1>(fn: (context: ContextValue<Context<C & V>, Revalue>, ...input: Input) => Revalue_1): {
                        readonly fn: (...input: Input) => Revalue_1;
                        pipe<Revalue_2>(fn: (context: ContextValue<Context<C & V>, Revalue_1>, ...input: Input) => Revalue_2): {
                            readonly fn: (...input: Input) => Revalue_2;
                            pipe<Revalue_3>(fn: (context: ContextValue<Context<C & V>, Revalue_2>, ...input: Input) => Revalue_3): {
                                readonly fn: (...input: Input) => Revalue_3;
                                pipe<Revalue_4>(fn: (context: ContextValue<Context<C & V>, Revalue_3>, ...input: Input) => Revalue_4): {
                                    readonly fn: (...input: Input) => Revalue_4;
                                    pipe<Revalue_5>(fn: (context: ContextValue<Context<C & V>, Revalue_4>, ...input: Input) => Revalue_5): {
                                        readonly fn: (...input: Input) => Revalue_5;
                                        pipe<Revalue_6>(fn: (context: ContextValue<Context<C & V>, Revalue_5>, ...input: Input) => Revalue_6): {
                                            readonly fn: (...input: Input) => Revalue_6;
                                            pipe<Revalue_7>(fn: (context: ContextValue<Context<C & V>, Revalue_6>, ...input: Input) => Revalue_7): {
                                                readonly fn: (...input: Input) => Revalue_7;
                                                pipe<Revalue_8>(fn: (context: ContextValue<Context<C & V>, Revalue_7>, ...input: Input) => Revalue_8): {
                                                    readonly fn: (...input: Input) => Revalue_8;
                                                    pipe<Revalue_9>(fn: (context: ContextValue<Context<C & V>, Revalue_8>, ...input: Input) => Revalue_9): {
                                                        readonly fn: (...input: Input) => Revalue_9;
                                                        pipe<Revalue_10>(fn: (context: ContextValue<Context<C & V>, Revalue_9>, ...input: Input) => Revalue_10): any;
                                                    };
                                                };
                                            };
                                        };
                                    };
                                };
                            };
                        };
                    };
                };
            };
            permit: (scopes: string[]) => {
                fn: (...input: Input) => Value;
                pipe<Value_1>(func: (context: ContextValue<Context<C & V>, Value>, ...input: Input) => Value_1): {
                    readonly fn: (...input: Input) => Value_1;
                    pipe<Revalue>(fn: (context: ContextValue<Context<C & V>, Value_1>, ...input: Input) => Revalue): {
                        readonly fn: (...input: Input) => Revalue;
                        pipe<Revalue_1>(fn: (context: ContextValue<Context<C & V>, Revalue>, ...input: Input) => Revalue_1): {
                            readonly fn: (...input: Input) => Revalue_1;
                            pipe<Revalue_2>(fn: (context: ContextValue<Context<C & V>, Revalue_1>, ...input: Input) => Revalue_2): {
                                readonly fn: (...input: Input) => Revalue_2;
                                pipe<Revalue_3>(fn: (context: ContextValue<Context<C & V>, Revalue_2>, ...input: Input) => Revalue_3): {
                                    readonly fn: (...input: Input) => Revalue_3;
                                    pipe<Revalue_4>(fn: (context: ContextValue<Context<C & V>, Revalue_3>, ...input: Input) => Revalue_4): {
                                        readonly fn: (...input: Input) => Revalue_4;
                                        pipe<Revalue_5>(fn: (context: ContextValue<Context<C & V>, Revalue_4>, ...input: Input) => Revalue_5): {
                                            readonly fn: (...input: Input) => Revalue_5;
                                            pipe<Revalue_6>(fn: (context: ContextValue<Context<C & V>, Revalue_5>, ...input: Input) => Revalue_6): {
                                                readonly fn: (...input: Input) => Revalue_6;
                                                pipe<Revalue_7>(fn: (context: ContextValue<Context<C & V>, Revalue_6>, ...input: Input) => Revalue_7): {
                                                    readonly fn: (...input: Input) => Revalue_7;
                                                    pipe<Revalue_8>(fn: (context: ContextValue<Context<C & V>, Revalue_7>, ...input: Input) => Revalue_8): {
                                                        readonly fn: (...input: Input) => Revalue_8;
                                                        pipe<Revalue_9>(fn: (context: ContextValue<Context<C & V>, Revalue_8>, ...input: Input) => Revalue_9): {
                                                            readonly fn: (...input: Input) => Revalue_9;
                                                            pipe<Revalue_10>(fn: (context: ContextValue<Context<C & V>, Revalue_9>, ...input: Input) => Revalue_10): any;
                                                        };
                                                    };
                                                };
                                            };
                                        };
                                    };
                                };
                            };
                        };
                    };
                };
            };
        };
    };
    protect<S_1 extends Schema>(schema: AccessSchema<S_1>, guard: (permit: {
        get: <Prefix_2 extends string = "", Spacer_2 extends string = "">(scope: UnionScopes<S_1, Prefix_2, Spacer_2>) => PermitNode | undefined;
        has: <Prefix_3 extends string = "", Spacer_3 extends string = "">(scope: UnionScopes<S_1, Prefix_3, Spacer_3>) => boolean;
        search: (substring: string, handle: (node: PermitNode) => void) => void;
        getChild: (scope: string) => PermitNode[];
        reset: () => void;
        grant: (...scopes: string[] | SchemaNode[]) => void;
        deny: (...scopes: string[] | SchemaNode[]) => void;
        check(scope: string): {
            readonly exact: boolean;
            readonly hasChild: boolean;
            readonly hasAnyChildGranted: boolean;
            readonly hasAllChildGranted: boolean;
            readonly hasAllChildDenied: boolean;
        };
        readonly export: {
            readonly all: {
                scope: string;
                desc: string;
            }[];
            readonly granted: {
                scope: string;
                desc: string;
            }[];
            readonly denied: {
                scope: string;
                desc: string;
            }[];
        };
        readonly granted: {
            scope: string;
            desc: string;
        }[];
        readonly denied: {
            scope: string;
            desc: string;
        }[];
        readonly nodes: TransformSchema<S_1, PermitNode>;
        readonly can: TransformSchema<S_1, boolean>;
    }) => boolean): {
        procedure: <Input extends any[], Value = void>(fn?: (context: Context<C & V>, ...args: Input) => Value) => {
            fn: (...input: Input) => Value;
            pipe<Value_1>(func: (context: ContextValue<Context<C & V>, Value>, ...input: Input) => Value_1): {
                readonly fn: (...input: Input) => Value_1;
                pipe<Revalue>(fn: (context: ContextValue<Context<C & V>, Value_1>, ...input: Input) => Revalue): {
                    readonly fn: (...input: Input) => Revalue;
                    pipe<Revalue_1>(fn: (context: ContextValue<Context<C & V>, Revalue>, ...input: Input) => Revalue_1): {
                        readonly fn: (...input: Input) => Revalue_1;
                        pipe<Revalue_2>(fn: (context: ContextValue<Context<C & V>, Revalue_1>, ...input: Input) => Revalue_2): {
                            readonly fn: (...input: Input) => Revalue_2;
                            pipe<Revalue_3>(fn: (context: ContextValue<Context<C & V>, Revalue_2>, ...input: Input) => Revalue_3): {
                                readonly fn: (...input: Input) => Revalue_3;
                                pipe<Revalue_4>(fn: (context: ContextValue<Context<C & V>, Revalue_3>, ...input: Input) => Revalue_4): {
                                    readonly fn: (...input: Input) => Revalue_4;
                                    pipe<Revalue_5>(fn: (context: ContextValue<Context<C & V>, Revalue_4>, ...input: Input) => Revalue_5): {
                                        readonly fn: (...input: Input) => Revalue_5;
                                        pipe<Revalue_6>(fn: (context: ContextValue<Context<C & V>, Revalue_5>, ...input: Input) => Revalue_6): {
                                            readonly fn: (...input: Input) => Revalue_6;
                                            pipe<Revalue_7>(fn: (context: ContextValue<Context<C & V>, Revalue_6>, ...input: Input) => Revalue_7): {
                                                readonly fn: (...input: Input) => Revalue_7;
                                                pipe<Revalue_8>(fn: (context: ContextValue<Context<C & V>, Revalue_7>, ...input: Input) => Revalue_8): {
                                                    readonly fn: (...input: Input) => Revalue_8;
                                                    pipe<Revalue_9>(fn: (context: ContextValue<Context<C & V>, Revalue_8>, ...input: Input) => Revalue_9): {
                                                        readonly fn: (...input: Input) => Revalue_9;
                                                        pipe<Revalue_10>(fn: (context: ContextValue<Context<C & V>, Revalue_9>, ...input: Input) => Revalue_10): any;
                                                    };
                                                };
                                            };
                                        };
                                    };
                                };
                            };
                        };
                    };
                };
            };
            protect<S extends Schema>(schema: AccessSchema<S>, guard: (permit: {
                get: <Prefix extends string = "", Spacer extends string = "">(scope: UnionScopes<S, Prefix, Spacer>) => PermitNode | undefined;
                has: <Prefix_1 extends string = "", Spacer_1 extends string = "">(scope: UnionScopes<S, Prefix_1, Spacer_1>) => boolean;
                search: (substring: string, handle: (node: PermitNode) => void) => void;
                getChild: (scope: string) => PermitNode[];
                reset: () => void;
                grant: (...scopes: string[] | SchemaNode[]) => void;
                deny: (...scopes: string[] | SchemaNode[]) => void;
                check(scope: string): {
                    readonly exact: boolean;
                    readonly hasChild: boolean;
                    readonly hasAnyChildGranted: boolean;
                    readonly hasAllChildGranted: boolean;
                    readonly hasAllChildDenied: boolean;
                };
                readonly export: {
                    readonly all: {
                        scope: string;
                        desc: string;
                    }[];
                    readonly granted: {
                        scope: string;
                        desc: string;
                    }[];
                    readonly denied: {
                        scope: string;
                        desc: string;
                    }[];
                };
                readonly granted: {
                    scope: string;
                    desc: string;
                }[];
                readonly denied: {
                    scope: string;
                    desc: string;
                }[];
                readonly nodes: TransformSchema<S, PermitNode>;
                readonly can: TransformSchema<S, boolean>;
            }) => boolean): {
                fn: (...input: Input) => Value;
                pipe<Value_1>(func: (context: ContextValue<Context<C & V>, Value>, ...input: Input) => Value_1): {
                    readonly fn: (...input: Input) => Value_1;
                    pipe<Revalue>(fn: (context: ContextValue<Context<C & V>, Value_1>, ...input: Input) => Revalue): {
                        readonly fn: (...input: Input) => Revalue;
                        pipe<Revalue_1>(fn: (context: ContextValue<Context<C & V>, Revalue>, ...input: Input) => Revalue_1): {
                            readonly fn: (...input: Input) => Revalue_1;
                            pipe<Revalue_2>(fn: (context: ContextValue<Context<C & V>, Revalue_1>, ...input: Input) => Revalue_2): {
                                readonly fn: (...input: Input) => Revalue_2;
                                pipe<Revalue_3>(fn: (context: ContextValue<Context<C & V>, Revalue_2>, ...input: Input) => Revalue_3): {
                                    readonly fn: (...input: Input) => Revalue_3;
                                    pipe<Revalue_4>(fn: (context: ContextValue<Context<C & V>, Revalue_3>, ...input: Input) => Revalue_4): {
                                        readonly fn: (...input: Input) => Revalue_4;
                                        pipe<Revalue_5>(fn: (context: ContextValue<Context<C & V>, Revalue_4>, ...input: Input) => Revalue_5): {
                                            readonly fn: (...input: Input) => Revalue_5;
                                            pipe<Revalue_6>(fn: (context: ContextValue<Context<C & V>, Revalue_5>, ...input: Input) => Revalue_6): {
                                                readonly fn: (...input: Input) => Revalue_6;
                                                pipe<Revalue_7>(fn: (context: ContextValue<Context<C & V>, Revalue_6>, ...input: Input) => Revalue_7): {
                                                    readonly fn: (...input: Input) => Revalue_7;
                                                    pipe<Revalue_8>(fn: (context: ContextValue<Context<C & V>, Revalue_7>, ...input: Input) => Revalue_8): {
                                                        readonly fn: (...input: Input) => Revalue_8;
                                                        pipe<Revalue_9>(fn: (context: ContextValue<Context<C & V>, Revalue_8>, ...input: Input) => Revalue_9): {
                                                            readonly fn: (...input: Input) => Revalue_9;
                                                            pipe<Revalue_10>(fn: (context: ContextValue<Context<C & V>, Revalue_9>, ...input: Input) => Revalue_10): any;
                                                        };
                                                    };
                                                };
                                            };
                                        };
                                    };
                                };
                            };
                        };
                    };
                };
                permit: (scopes: string[]) => {
                    fn: (...input: Input) => Value;
                    pipe<Value_1>(func: (context: ContextValue<Context<C & V>, Value>, ...input: Input) => Value_1): {
                        readonly fn: (...input: Input) => Value_1;
                        pipe<Revalue>(fn: (context: ContextValue<Context<C & V>, Value_1>, ...input: Input) => Revalue): {
                            readonly fn: (...input: Input) => Revalue;
                            pipe<Revalue_1>(fn: (context: ContextValue<Context<C & V>, Revalue>, ...input: Input) => Revalue_1): {
                                readonly fn: (...input: Input) => Revalue_1;
                                pipe<Revalue_2>(fn: (context: ContextValue<Context<C & V>, Revalue_1>, ...input: Input) => Revalue_2): {
                                    readonly fn: (...input: Input) => Revalue_2;
                                    pipe<Revalue_3>(fn: (context: ContextValue<Context<C & V>, Revalue_2>, ...input: Input) => Revalue_3): {
                                        readonly fn: (...input: Input) => Revalue_3;
                                        pipe<Revalue_4>(fn: (context: ContextValue<Context<C & V>, Revalue_3>, ...input: Input) => Revalue_4): {
                                            readonly fn: (...input: Input) => Revalue_4;
                                            pipe<Revalue_5>(fn: (context: ContextValue<Context<C & V>, Revalue_4>, ...input: Input) => Revalue_5): {
                                                readonly fn: (...input: Input) => Revalue_5;
                                                pipe<Revalue_6>(fn: (context: ContextValue<Context<C & V>, Revalue_5>, ...input: Input) => Revalue_6): {
                                                    readonly fn: (...input: Input) => Revalue_6;
                                                    pipe<Revalue_7>(fn: (context: ContextValue<Context<C & V>, Revalue_6>, ...input: Input) => Revalue_7): {
                                                        readonly fn: (...input: Input) => Revalue_7;
                                                        pipe<Revalue_8>(fn: (context: ContextValue<Context<C & V>, Revalue_7>, ...input: Input) => Revalue_8): {
                                                            readonly fn: (...input: Input) => Revalue_8;
                                                            pipe<Revalue_9>(fn: (context: ContextValue<Context<C & V>, Revalue_8>, ...input: Input) => Revalue_9): {
                                                                readonly fn: (...input: Input) => Revalue_9;
                                                                pipe<Revalue_10>(fn: (context: ContextValue<Context<C & V>, Revalue_9>, ...input: Input) => Revalue_10): any;
                                                            };
                                                        };
                                                    };
                                                };
                                            };
                                        };
                                    };
                                };
                            };
                        };
                    };
                };
            };
        };
    };
};
declare function inputSchema<T, Value>(schema: ZodType<Value>): (_: Context<T>, input: Value) => Value;

declare const permit: {
    createPermit: typeof createPermit;
    createSchema: typeof createSchema;
    createGuard: typeof createGuard;
};
declare const garda: {
    createProcedure: typeof createProcedure;
    createInterface: typeof createInterface;
    createContext: typeof createContext;
    inputSchema: typeof inputSchema;
};

export { DataRecord, DataType, Inferer, ZodType, garda, infer, permit };
