/* eslint-disable @typescript-eslint/no-explicit-any */

import { ZodType } from '../infer';
import { AccessSchema, Permit, Schema } from '../permit';

type ContextRoot = { scopes?: string[]; protected?: boolean };
type Context<T> = T & ContextRoot;
type ContextValue<T, V> = T & { value: V } & ContextRoot;

export function createProcedure<
  T extends Record<string, any>,
  I extends any[],
  V = undefined
>(
  ctx: () => Context<T>,
  fn: (context: Context<T>, ...args: I) => V = () => undefined as V
) {
  type PipeFunction<Input, Output> = (
    context: ContextValue<T, Input>,
    ...input: I
  ) => Output;
  type EndFunction<Value> = (...input: I) => {
    value: Value;
    context: Context<T>;
  };

  const recursivePipeGenerator = <NextValue, Value>(
    func: PipeFunction<Value, NextValue>,
    endFn: EndFunction<Value>
  ) => {
    return {
      // generate end function
      get fn() {
        return (...input: I) => {
          const { context, value } = endFn(...input);
          return func({ ...context, value }, ...input);
        };
      },

      // generate new pipe function recursively
      pipe<Revalue>(fn: PipeFunction<NextValue, Revalue>) {
        return recursivePipeGenerator<Revalue, NextValue>(fn, (...input: I) => {
          const { context, value } = endFn(...input);
          return { context, value: func({ ...context, value }, ...input) };
        });
      },
    };
  };

  const beginPipeGenerator = (
    injector: () => Context<T>,
    extras: Record<string, any> = {},
    fnHook: ((context: Context<T>) => Context<T>) | undefined = undefined
  ) => {
    return {
      get fn() {
        return (...input: I) => {
          let _context = injector();
          if (typeof fnHook === 'function') {
            _context = fnHook(_context);
          }

          return fn(_context, ...input);
        };
      },
      pipe<Value>(func: PipeFunction<V, Value>) {
        return recursivePipeGenerator<Value, V>(func, (...input: I) => {
          const context = injector();
          const value = fn(context, ...input) as V;
          return { context, value };
        });
      },
      ...extras,
    };
  };

  return {
    protect<S extends Schema>(
      schema: AccessSchema<S>,
      guard: (permit: Permit<S>) => boolean
    ) {
      const joiningScopes = (a: string[], b: string[]) => {
        const joinedScopes = [...a, ...b];
        return [...new Set(joinedScopes)];
      };

      const permitContextInjector = (scopes: string[]) => {
        return () => {
          const context = ctx();
          return {
            ...context,
            scopes: joiningScopes(context.scopes ?? [], scopes),
          };
        };
      };

      const guardEndFunction = (context: Context<T>) => {
        const _scopes = context.scopes ?? [];
        if (_scopes.length !== 0) {
          const permit = schema.permit(_scopes);
          context.protected = context.protected && guard(permit as Permit<S>);
        }
        return context;
      };

      const permitFn = (scopes: string[]) => {
        const injector = permitContextInjector(scopes);
        return {
          ...beginPipeGenerator(
            injector,
            {
              permit: permitFn,
            },
            guardEndFunction
          ),
        };
      };

      return {
        // generate fn, pipe, permit
        permit: permitFn,

        // generate fn, pipe
        ...beginPipeGenerator(permitContextInjector([]), {}, guardEndFunction),
      };
    },

    // generate fn and pipe
    ...beginPipeGenerator(ctx),
  };
}

export function createContext<T>(ctx: () => Context<T>, immidiate = false) {
  // if immidiate call is false, return context function as is
  if (!immidiate) return ctx;

  // run context immidiately
  // then return a function that return context result.
  const context = ctx();
  return () => context;
}

export type InterfaceConfig<TC, SC> = {
  sharedContext: () => Context<SC>;
  context: () => Context<TC>;
  onError?: (error: Error) => void;
  onCall?: (value: any) => void;
};

export function createInterface<C, V>(
  config: Partial<InterfaceConfig<C, V>> = {}
) {
  const _config: InterfaceConfig<C, V> = {
    sharedContext: () => ({} as Context<V>),
    context: () => ({} as Context<C>),
    ...config,
  };

  type T = Context<C & V>;
  const sharedContext = _config.sharedContext();
  const ctx: () => T = () => {
    return {
      ..._config.context(),
      ...sharedContext,
    } as T;
  };

  const procedureImplementor = (injector: () => T) => {
    return <Input extends any[], Value = void>(
      fn: (context: T, ...args: Input) => Value = () => undefined as Value
    ) => {
      return createProcedure<T, Input, Value>(injector, fn);
    };
  };

  return {
    procedure: procedureImplementor(ctx),
    protect<S extends Schema>(
      schema: AccessSchema<S>,
      guard: (permit: Permit<S>) => boolean
    ) {
      const protectedContext = () => {
        const context = ctx();
        const permit = schema.permit(context.scopes ?? []);
        context.protected = guard(permit as Permit<S>);
        return context;
      };

      return { procedure: procedureImplementor(protectedContext) };
    },
  };
}

export function inputSchema<T, Value>(schema: ZodType<Value>) {
  return (_: Context<T>, input: Value) => {
    return schema.parse(input);
  };
}

export function outputSchema<T, Value>(schema: ZodType<Value>) {
  return (context: ContextValue<T, any>) => {
    return schema.parse(context.value);
  };
}