import { describe, expect, test } from 'vitest';
import {
  createContext,
  createInterface,
  createProcedure,
  inputSchema,
  outputSchema,
} from '..';
import { z } from 'zod';
import { createSchema } from '../../permit';

const schema = {
  secret: {
    read: 'read value of secret',
    write: 'update value of secret',
  },
};

const access = createSchema(schema);

describe('Garda Unit Testing', () => {
  const $public = createInterface({
    context() {
      return {
        logs(context: string, ...tags: string[]) {
          return { context, tags };
        },
      };
    },
    sharedContext: () => {
      let count = 0;
      const counter = (add = 1) => (count = count + add);

      return {
        counter,
        set count(num: number) {
          count = num;
        },
        get count() {
          return count;
        },
      };
    },
  });

  test('Garda Open API', () => {
    const setCounter = $public.procedure((context, counter: number) => {
      context.count = counter;
    }).fn;

    const counterProc = $public.procedure(({ count }) => count);
    const getCounterNumber = counterProc.fn;
    const getCounterString = counterProc.pipe(({ value }) => value + '').fn;
    const addCounter = $public
      .procedure((_, count: number) => count)
      .pipe(({ counter, value }) => counter(value)).fn;

    setCounter(0);
    expect(getCounterNumber()).toBe(0);
    expect(getCounterString()).toBe('0');

    addCounter(2);
    expect(getCounterNumber()).toBe(2);
    expect(getCounterString()).toBe('2');
  });

  const $private = createInterface({
    sharedContext() {
      let _secret = 'some secret';
      const secret = {
        get value() {
          return _secret;
        },
        set value(secret: string) {
          _secret = secret;
        },
      };

      return {
        secret,
        scopes: ['secret.read'],
        verify(password: string) {
          return password === _secret;
        },
      };
    },
  }).protect<typeof schema>(
    access,
    (p) => p.check('secret').hasAnyChildGranted
  );

  test('Garda Private API', () => {
    const setSecret = $private
      .procedure<[secret: string, log?: string]>(
        (context, secret: string, log = '') => {
          if (log !== '') console.log(log);
          if (context.protected === true) context.secret.value = secret;
        }
      )
      .protect<typeof schema>(access, (p) => p.can.secret.write);

    const getSecret = $private
      .procedure((ctx) => ctx.secret.value)
      .protect<typeof schema>(access, (p) => p.can.secret.read);

    const verifySecret = $private.procedure((ctx, secret: string) =>
      ctx.verify(secret)
    );

    expect(getSecret.fn()).toBe('some secret');
    const secret = 'another secret string only you could know';

    // current state only able to secret.read

    setSecret.fn(secret);
    expect(getSecret.fn()).toBe('some secret');
    expect(verifySecret.fn(secret)).toBeFalsy();

    // now set able to secret.write
    setSecret.permit(['secret.write']).fn(secret);
    expect(getSecret.fn()).toBe(secret);
    expect(verifySecret.fn(secret)).toBeTruthy();

    // previous set secret only able to set once
    setSecret.fn('another secret');
    expect(getSecret.fn()).toBe(secret);

    // try create openSetSecret that can always set secret
    const openSetSecret = setSecret.permit(['secret.write']).fn;
    openSetSecret('nice');
    expect(getSecret.fn()).toBe('nice');
    openSetSecret('welldone');
    expect(getSecret.fn()).toBe('welldone');
  });

  test('Garda Utility Function', () => {
    // setup context
    const counterContext = createContext(() => {
      let count = 0;

      return {
        count: {
          add: (value = 1) => {
            count += value;
          },
          set value(num: number) {
            count = num;
          },
          get value() {
            return count;
          },
        },
      };
    }, true);

    // restart counter fn
    const restartCounter = createProcedure(
      counterContext,
      (ctx) => (ctx.count.value = 0)
    ).fn;

    // increment counter fn
    const incrementCount = createProcedure(
      counterContext,
      inputSchema(z.number())
    ).pipe(({ value, count }) => count.add(value)).fn;

    const getCountValue = createProcedure(
      counterContext,
      (ctx) => ctx.count.value
    );

    // get current count number fn
    const getCountNumber = getCountValue.fn;

    const getCountString = getCountValue
      .pipe(({ value }) => value + '')
      .pipe(outputSchema(z.string())).fn;

    const multiplyCount = createProcedure(
      counterContext,
      inputSchema(z.number())
    )
      .pipe(({ value, count }) => {
        const prevCount = count.value;
        count.value = count.value * value;
        return {
          result: count.value,
          verbose: `${prevCount}*${value}=${count.value}`,
        };
      })
      .pipe(
        outputSchema(
          z.object({
            result: z.number(),
            verbose: z.string(),
          })
        )
      ).fn;

    // for start counter should be 0
    expect(getCountNumber()).toBe(0);

    // increment counter to 2
    incrementCount(2);
    expect(getCountNumber()).toBe(2);

    // restart counter to 0
    restartCounter();
    expect(getCountNumber()).toBe(0);

    expect(getCountString()).toBeTypeOf('string');
    expect(getCountString()).toBe('0');

    // test multiplying 2 by 2
    incrementCount(2);
    const multiplied = multiplyCount(2);
    expect(multiplied.result).toBe(4);
    expect(multiplied.verbose).toBe('2*2=4');
    expect(getCountNumber()).toBe(4);
  });
});
