/* eslint-disable @typescript-eslint/no-explicit-any */
export type ZodType<T> = {
  parse: (data: any) => T;
  array: () => ZodType<T[]>;
};

type parseOptions<T> = {
  default?: T;
  parseHandler?: (value: DataType) => T;
};

export type DataRecord = { [P in string]: DataType };
export type DataType =
  | DataRecord
  | DataRecord[]
  | any[]
  | string
  | number
  | boolean
  | null
  | undefined;

type Handler<T> = (row: Inferer) => T;

export type Inferer = ReturnType<typeof infer>;

export function infer(data: DataType) {
  /**
   * Get raw value based on object path
   *
   * @param {string} [path='']
   * @return {DataType}
   */
  function _get(path = '') {
    if (path === '.' || path === '') return data;
    if (!data || typeof data !== 'object') return data;
    if (data instanceof Array) return data;

    const paths = path.split('.');
    let value: DataType = undefined;
    for (let index = 0; index < paths.length; index++) {
      const key = paths[index];
      if (index === 0) {
        if (Object.hasOwn(data, key)) {
          value = data[key];
          continue;
        }
      } else {
        if (value && typeof value === 'object' && !(value instanceof Array)) {
          if (Object.hasOwn(value, key)) {
            value = value[key];
            continue;
          }
        }
      }
    }

    return value;
  }

  function createError(...tags: string[]) {
    return new Error(['[Infer Error]', ...tags].join(' '));
  }

  function conversion<T>(
    path: string,
    type: string,
    value: DataType,
    options: parseOptions<T>
  ) {
    if (options.parseHandler) return options.parseHandler(value);
    if (options.default) return options.default ?? '';
    throw createError(
      `[path: '${path}']`,
      `[type: ${type}]`,
      `[value: ${value}]`
    );
  }

  /**
   * Infer object path in runtime either value is string or throw an error
   *
   * *NOTE - path doesn't work with array. 'foo.bar[0]' this will not work. use v(data).array('foo.bar') to validate array.
   *
   * @param {string} path
   * @param {parseOptions<string>} options
   * @return {string}
   */
  function string(path = '', options: parseOptions<string> = {}) {
    const value = _get(path);
    if (typeof value === 'string') return value;
    return conversion(path, 'string', value, options);
  }

  /**
   * Infer object path in runtime either value is number (int|float) or throw an error
   *
   * *NOTE - path doesn't work with array. 'foo.bar[0]' this will not work. use v(data).array('foo.bar') to validate array.
   *
   * @param {string} path
   * @param {(parseOptions<number> & { float?: boolean })} options
   * @return {number}
   */
  function number(
    path = '',
    options: parseOptions<number> & { float?: boolean } = {}
  ) {
    const value = _get(path);
    if (typeof value === 'number') return value;
    if (typeof value === 'string') {
      if (options.float === true) {
        const parsed = parseFloat(value);
        if (!isNaN(parsed)) return parsed;
      } else {
        const parsed = parseInt(value);
        if (!isNaN(parsed)) return parsed;
      }
    }
    return conversion(path, 'number', value, options);
  }

  /**
   * Infer object path in runtime either value is boolean or throw an error
   *
   * *NOTE - path doesn't work with array. 'foo.bar[0]' this will not work. use v(data).array('foo.bar') to validate array.
   *
   * @param {string} path
   * @param {(parseOptions<number> & { float?: boolean })} options
   * @return {number}
   */
  function boolean(path = '', options: parseOptions<boolean> = {}) {
    const value = _get(path);
    if (typeof value === 'boolean') return value;
    return conversion(path, 'boolean', value, options);
  }

  /**
   * Infer object path in runtime either value is string|number or throw an error
   *
   * *NOTE - path doesn't work with array. 'foo.bar[0]' this will not work. use v(data).array('foo.bar') to validate array.
   *
   * @param {string} path
   * @param {(parseOptions<string | number> & { float?: boolean })} options
   * @return {string|number}
   */
  function stringOrNumber(
    path = '',
    options: parseOptions<string | number> & { float?: boolean } = {}
  ) {
    const value = _get(path);
    if (typeof value === 'string') return value;
    if (typeof value === 'number') return value;
    return conversion(path, 'string|number', value, options);
  }

  /**
   * Infer object path in runtime is array of given zod schema or zod type parse
   *
   * @template T
   * @param {string} path
   * @param {ZodType<T>} schema
   * @param {parseOptions<T[]>} options
   * @return {T[]}
   */
  function array<T = DataType>(
    path: string,
    schema: ZodType<T> | Handler<T>,
    options: parseOptions<T[]> = {}
  ) {
    const value = _get(path);
    if (value instanceof Array) {
      if (typeof schema === 'function') {
        const list: T[] = [];
        for (let i = 0; i < value.length; i++) {
          const row = value[i] as DataType;
          list.push(schema(infer(row)));
        }

        return list;
      } else {
        return schema.array().parse(value);
      }
    }
    return conversion(path, 'array', value, options);
  }

  /**
   * Infer object path in runtime using zod schema
   *
   * @template T
   * @param {string} path
   * @param {ZodType<T>} schema
   * @return {T}
   */
  function object<T>(
    path: string,
    schema: ZodType<T> | Handler<T>,
    options: parseOptions<T> = {}
  ): T {
    const value = _get(path);
    if (value && typeof value === 'object' && !(value instanceof Array)) {
      if (typeof schema === 'function') {
        return schema(infer(value));
      } else {
        return schema.parse(value);
      }
    }

    return conversion(path, 'object', value, options);
  }

  /**
   * Infer object path in runtime and transform data in array with given handler.
   *
   * @template F
   * @param {string} path
   * @param {F} handler
   * @return {ReturnType<F>}
   */
  function map<T>(path = '', handler: Handler<T>): T[] {
    const value = _get(path);
    if (value instanceof Array) {
      const list: T[] = [];
      for (let i = 0; i < value.length; i++) {
        const item = value[i] as DataRecord;
        if (item && typeof item === 'object' && !(item instanceof Array)) {
          list[i] = handler(infer(item));
        } else {
          throw createError(
            `[path: ${path}[${i}]]`,
            '[map: given path values is not array of records]'
          );
        }
      }

      return list;
    }
    throw createError(
      `[path:'${path}']`,
      '[map: given path values is not an array]'
    );
  }

  /**
   * Infer object path in runtime if value type wrong return null
   *
   * @param {string} path
   */
  function nullable(path = '') {
    const wrap = <A extends any[], Fn extends (...args: A) => any>(
      fn: Fn,
      ...args: A
    ) => {
      try {
        return fn(...args) as ReturnType<Fn>;
      } catch (error) {
        return null;
      }
    };

    return {
      string: (options: parseOptions<string> = {}) =>
        wrap(string, path, options),
      number: (options: parseOptions<number> & { float?: boolean } = {}) =>
        wrap(number, path, options),
      boolean: (options: parseOptions<boolean> = {}) =>
        wrap(boolean, path, options),
      stringOrNumber: (
        options: parseOptions<string | number> & { float?: boolean } = {}
      ) => wrap(stringOrNumber, path, options),
      array: <T = DataType>(
        handler: ZodType<T> | Handler<T>,
        options: parseOptions<T[]> = {}
      ) => {
        try {
          return array<T>(path, handler, options);
        } catch (error) {
          return null;
        }
      },
      object: <T>(
        handler: ZodType<T> | Handler<T>,
        options: parseOptions<T> = {}
      ) => {
        try {
          return object<T>(path, handler, options);
        } catch (error) {
          return null;
        }
      },
      map: <T>(handler: Handler<T>): null | T[] => {
        try {
          return map(path, handler);
        } catch (error) {
          return null;
        }
      },
    };
  }

  return {
    map,
    array,
    object,
    string,
    number,
    boolean,
    nullable,
    get: _get,
    stringOrNumber,
    get raw() {
      return data;
    },
    has(path: string) {
      const found = _get(path);
      return found !== undefined && found !== null;
    },
  };
}
