import { describe, expect, test } from 'vitest';
import { infer } from '../..';
import { ZodError, z } from 'zod';

describe('Infer data test', () => {
  const data = JSON.parse(
    JSON.stringify({
      records: {
        count: 3,
        items: [
          { id: 'one', value: 1, toggle: true },
          { id: 'two', value: 2, toggle: false },
          { id: 'three', value: 3, toggle: true },
        ],
        last: 'three',
      },
      bool: false,
      list: ['one', 'two', 'three'],
      float: '10.000000777',
      map: [
        { name: 'john' },
        { name: 'smith' },
        { name: 'burk' },
        { name: 'leila' },
      ],
    })
  );

  test('string infer success', () => {
    const v = infer(data);
    expect(v.string('records.last')).toBe('three');
  });

  test('string infer fail and throw error', () => {
    const v = infer(data);
    expect(() => v.string('records.count')).toThrowError('Infer Error');
    expect(() => v.string('records.count')).toThrowError('records.count');
    expect(() => v.string('records.count')).toThrowError('string');
  });

  test('number infer success', () => {
    const v = infer(data);
    expect(v.number('records.count')).toBe(3);
  });

  test('number infer fail and throw error', () => {
    const v = infer(data);
    expect(() => v.number('records.last')).toThrowError('Infer Error');
    expect(() => v.number('records.last')).toThrowError('records.last');
    expect(() => v.number('records.last')).toThrowError('number');
  });

  test('boolean infer success', () => {
    const v = infer(data);
    expect(v.boolean('bool')).toBe(false);
  });

  test('boolean infer fail and throw error', () => {
    const v = infer(data);
    expect(() => v.boolean('records')).toThrowError('Infer Error');
    expect(() => v.boolean('records')).toThrowError('records');
    expect(() => v.boolean('records')).toThrowError('boolean');
  });

  test('string or number infer success', () => {
    const v = infer(data);
    expect(v.stringOrNumber('float')).toBe('10.000000777');
  });

  test('string or number infer fail and throw error', () => {
    const v = infer(data);
    expect(() => v.stringOrNumber('list')).toThrowError('Infer Error');
    expect(() => v.stringOrNumber('list')).toThrowError('list');
    expect(() => v.stringOrNumber('list')).toThrowError('string|number');
  });

  test('array infer success', () => {
    const v = infer(data);
    const list = v.array(
      'records.items',
      z.object({ id: z.string(), value: z.number(), toggle: z.boolean() })
    );

    expect(list).toBeInstanceOf(Array);
    expect(list[0].id).toBe('one');
  });

  test('array infer fail and throw error', () => {
    const v = infer(data);
    expect(() => v.array('records', z.string())).toThrowError('Infer Error');
    expect(() => v.array('records', z.string())).toThrowError('records');
    expect(() => v.array('records', z.string())).toThrowError('array');
  });

  test('array infer fail and throw zod error', () => {
    const v = infer(data);
    expect(() => v.array('records.items', z.string())).toThrowError(ZodError);
  });

  test('object infer success', () => {
    const v = infer(data);
    const record = v.object(
      'records',
      z.object({ count: z.number(), last: z.string() })
    );
    expect(record.count).toBe(3);
    expect(record.last).toBe('three');
  });

  test('object infer using get in zod parse', () => {
    const v = infer(data);
    const record = z
      .object({ count: z.number(), last: z.string() })
      .parse(v.get('records'));
    expect(record.count).toBe(3);
    expect(record.last).toBe('three');
  });

  test('object infer fail and throw error', () => {
    const v = infer(data);
    const fn = () =>
      v.object('map', z.object({ count: z.number(), last: z.string() }));

    expect(fn).toThrowError('Infer Error');
    expect(fn).toThrowError('object');
    expect(fn).toThrowError('map');
  });

  test('infer map helper success', () => {
    const v = infer(data);
    const list = v.map('map', (row) => {
      return {
        id: Math.floor(Math.random() * 10),
        name: row.string('name'),
      };
    });

    expect(list.length).toBe(4);
    expect(list[0].name).toBe('john');
  });

  test('infer map helper failed and throw error not array', () => {
    const v = infer(data);
    const fn = () => v.map('float', (row) => row.number('.'));
    expect(fn).toThrowError('Infer Error');
    expect(fn).toThrowError('not an array');
  });

  test('infer map helper failed and throw error not array of records', () => {
    const v = infer(data);
    const fn = () => v.map('list', (row) => row.string('.'));
    expect(fn).toThrowError('Infer Error');
    expect(fn).toThrowError('not array of records');
  });

  test('infer nullable', () => {
    const v = infer({});
    expect(v.nullable('a').string()).toBeNull();
    expect(v.nullable('a').number()).toBeNull();
    expect(v.nullable('a').boolean()).toBeNull();
    expect(v.nullable('a').stringOrNumber()).toBeNull();
    expect(v.nullable('a').array(z.string())).toBeNull();
    expect(v.nullable('a').object(z.string())).toBeNull();
    expect(v.nullable('a').map((row) => row.string('b'))).toBeNull();
  });
});
