import { AccessSchema, SchemaNode } from './schema';
import { Schema, SchemaOptions, UnionScopes } from './types';
import type { Permit } from './permit';

/**
 * Create access control schema
 *
 * @param {T} schema
 * @param {Partial<SchemaOptions>} [options={}]
 * @return {AccessSchema} access schema
 */
export function createSchema<T extends Schema>(
  schema: T,
  options: Partial<SchemaOptions> = {}
) {
  return new AccessSchema(schema, options);
}

type PermitOptions = {
  granted: string[] | SchemaNode[];
  denied: string[] | SchemaNode[];
} & SchemaOptions;

/**
 * Create user permit
 *
 * @export
 * @template T
 * @param {AccessSchema<T>} schema
 * @param {(string[] | SchemaNode[])} scopes
 * @return {Permit}
 */
export function createPermit<T extends Schema>(
  schema: T,
  options: Partial<PermitOptions> = {}
) {
  const schemaObject = new AccessSchema(schema, options);
  const permit = schemaObject.permit(options.granted ?? []);
  if (options.denied instanceof Array) {
    permit.deny(...options.denied);
  }

  return permit;
}

type SchemaType<T> = T extends AccessSchema<infer S> ? S : never;

/**
 * Create guard utility that check based on given guard function
 *
 * @export
 * @template T
 * @template P
 * @param {AccessSchema<T>} schema
 * @param {(permit: Permit<T, P>) => boolean} guard
 * @return {*} 
 */
export function createGuard<T extends Schema>(
  schema: AccessSchema<T>,
  guard: (permit: Permit<T>) => boolean
) {
  const permit = schema.permit([]) as Permit<T>;
  return {
    grant(scopes: string[]) {
      permit.grant(...scopes);
    },
    deny(scopes: string[]) {
      permit.deny(...scopes);
    },
    get isAllowed() {
      return guard(permit);
    },
    get isDenied() {
      return !guard(permit);
    },
  };
}

export type {
  AccessSchema,
  Schema,
  Permit,
  SchemaType,
  SchemaNode,
  SchemaOptions,
  UnionScopes,
};
