import { AccessSchema, SchemaNode } from './schema';
import { Schema, TransformSchema, PermitNode, UnionScopes } from './types';

type PermitScope = Omit<PermitNode, 'toggle' | 'granted'>;
export type Permit<T extends Schema> = ReturnType<typeof permit<T>>;

export function permit<T extends Schema>(
  access: AccessSchema<T>,
  values: string[]
) {
  /** stores */
  const _scopes = new Map<string, PermitScope>();
  const _values = new Map<string, boolean>();

  /** format _scopes to array */
  const _getScopesList = () => Array.from(_scopes.values());

  /** format _values to array */
  const _getValuesList = () => Array.from(_values.entries());

  /** utility for creating permit  */
  const _createPermitScope = (scope: string, desc: string): PermitScope => {
    _values.set(scope, values.includes(scope));
    return { scope, desc };
  };

  const _buildPermitNode = (node: PermitScope): PermitNode => ({
    ...node,
    get granted() {
      return _values.get(node.scope) ?? false;
    },
    set granted(value: boolean) {
      _values.set(node.scope, value);
    },
    toggle() {
      const current = _values.get(node.scope) ?? false;
      _values.set(node.scope, !current);
    },
  });

  const _schemaNodeExtract = (scope: string | SchemaNode) => {
    if (typeof scope === 'string') return scope;
    else return scope.scope;
  };

  /** reset scopes to default state */
  const reset = () => {
    access.export().forEach((scope) => {
      _scopes.set(
        scope.scope,
        _createPermitScope(scope.scope, scope.description)
      );
    });
  };

  /** grant access by string scopes */
  const grant = (...scopes: string[] | SchemaNode[]) => {
    for (let i = 0; i < scopes.length; i++) {
      const scope = scopes[i];
      if (_scopes.has(_schemaNodeExtract(scope))) {
        _values.set(_schemaNodeExtract(scope), true);
      }
    }
  };

  /** deny access by string scope */
  const deny = (...scopes: string[] | SchemaNode[]) => {
    for (let i = 0; i < scopes.length; i++) {
      const scope = scopes[i];
      if (_scopes.has(_schemaNodeExtract(scope))) {
        _values.set(_schemaNodeExtract(scope), false);
      }
    }
  };

  /** export getter builder */
  const initExportGetter = () => {
    const scopes = _getScopesList();
    const checkScopeValues = (node: PermitScope, value: boolean) =>
      _values.get(node.scope) === value;

    return {
      get all() {
        return scopes;
      },
      get granted() {
        return scopes.filter((node) => checkScopeValues(node, true));
      },
      get denied() {
        return scopes.filter((node) => checkScopeValues(node, false));
      },
    };
  };

  /** get access node by string scope */
  function get<Prefix extends string = '', Spacer extends string = ''>(
    scope: UnionScopes<T, Prefix, Spacer>
  ) {
    const node = _scopes.get(scope);
    if (!node) return undefined;
    return _buildPermitNode(node);
  }

  /** utility function for checking if there is an access node that have an exact match with given scope */
  function has<Prefix extends string = '', Spacer extends string = ''>(
    scope: UnionScopes<T, Prefix, Spacer>
  ) {
    const found = get(scope);
    return found !== undefined;
  }

  /** find access node that has the given substring. */
  function search(substring: string, handle: (node: PermitNode) => void) {
    const nodes = _getScopesList();
    for (let i = 0; i < nodes.length; i++) {
      const node = nodes[i];
      if (node.scope.toLowerCase().includes(substring.toLowerCase())) {
        handle(_buildPermitNode(node));
      }
    }
  }

  /** find all child of the given string scope */
  function getChild(scope: string) {
    const scopes = [];
    const nodes = _getScopesList();
    for (let i = 0; i < nodes.length; i++) {
      if (nodes[i].scope.startsWith(scope)) {
        scopes.push(_buildPermitNode(nodes[i]));
      }
    }

    return scopes;
  }

  // fresh permit from scopes and default
  reset();

  return {
    get,
    has,
    search,
    getChild,
    reset,
    grant,
    deny,

    /** utility for checking string scope permit value */
    check(scope: string) {
      const exact = _values.get(scope);
      const checkChild = (
        checker: (current: boolean, next: boolean) => boolean,
        starter = false
      ) => {
        // get only child values
        const nodes = _getValuesList().filter(([_scope]) =>
          _scope.startsWith(scope)
        );

        if (nodes.length === 0) return false;

        return nodes.reduce((result, [_, value]) => {
          return checker(value, result);
        }, starter);
      };

      return {
        get exact() {
          return exact ?? false;
        },
        get hasChild() {
          return exact === undefined;
        },
        get hasAnyChildGranted() {
          if (typeof exact === 'boolean') return exact;
          return checkChild((current, next) => current || next, false);
        },
        get hasAllChildGranted() {
          if (typeof exact === 'boolean') return exact;
          return checkChild((current, next) => current && next, true);
        },
        get hasAllChildDenied() {
          if (typeof exact === 'boolean') return exact;
          return checkChild((current, next) => !current && next, false);
        },
      };
    },

    /** getter for exporting scopes */
    get export() {
      return initExportGetter();
    },

    /** readonly enabled access node */
    get granted() {
      return _getScopesList().filter(
        (node) => _values.get(node.scope) === true
      );
    },

    /** readonly disabled access node */
    get denied() {
      return _getScopesList().filter(
        (node) => _values.get(node.scope) === false
      );
    },

    /** readonly AccessNodes  */
    get nodes(): TransformSchema<T, PermitNode> {
      return access.map((node) => {
        const found = _scopes.get(node.scope);
        if (found !== undefined) return _buildPermitNode(found);
        const scope = _createPermitScope(node.scope, node.description);
        _scopes.set(node.scope, scope);
        return _buildPermitNode(scope);
      });
    },

    /** readonly assertion utility for granted scopes  */
    get can(): TransformSchema<T, boolean> {
      return access.map((node) => _values.get(node.scope) ?? false);
    },
  };
}
