import { permit } from './permit';
import type { TransformSchema, Schema, SchemaOptions } from './types';
import { transformAccessControl, traverseAccessControl } from './utils';

export type SchemaNode = {
  scope: string;
  description: string;
};

export class AccessSchema<T extends Schema = Schema> {
  private _options: SchemaOptions;
  private _schema: T;
  private _scopes: Map<string, SchemaNode>;

  constructor(schema: T, options: Partial<SchemaOptions> = {}) {
    this._options = {
      prefix: options.prefix ?? '',
      spacer: options.spacer ?? '.',
    };

    this._schema = schema;
    this._scopes = new Map();
    this.resetSchemaNode();
  }

  private createSchemaNode(scope: string, description: string) {
    return { scope, description };
  }

  private resetSchemaNode() {
    traverseAccessControl(
      this._schema,
      this._options.prefix,
      this._options.spacer,
      (node, scope) => {
        this._scopes.set(scope, this.createSchemaNode(scope, node));
      }
    );
  }

  private get scopesList() {
    return Array.from(this._scopes.values());
  }

  /** utility function for finding scope */
  private _get(scope: string, opts: { exact: true }): SchemaNode | undefined;
  /** utility function for finding multiple scope */
  private _get(scope: string, opts: { exact: false }): SchemaNode[];
  /** scope utility function */
  private _get(scope: string, opts: { exact?: boolean } = {}) {
    if (opts.exact === true || opts.exact === undefined) {
      return this.scopesList.find((s) => s.scope === scope);
    }

    return this.scopesList.filter((s) => s.scope.startsWith(scope));
  }

  /** readonly schema options */
  get options() {
    return this._options;
  }

  /** readonly AccessNodes  */
  get nodes(): TransformSchema<T, string> {
    return transformAccessControl(
      structuredClone(this._schema),
      this._options.prefix,
      this._options.spacer,
      (desc) => desc
    ) as TransformSchema<T, string>;
  }

  /** export schema to list of scope and description */
  export() {
    return this.scopesList;
  }

  /** get schema node by string scope */
  get(scope: string) {
    return this._get(scope, { exact: true });
  }

  /** utility function for checking if there is a schema node that have an exact match with given scope */
  has(scope: string) {
    const found = this._get(scope, { exact: true });
    return found !== undefined;
  }

  /** find schema node that has the given substring in scope or description. */
  search(
    substring: string,
    fn = (n: SchemaNode) =>
      n.scope.includes(substring) || n.description.includes(substring)
  ) {
    return this.scopesList.filter(fn);
  }

  /** find all child of the given string scope */
  getChild(scope: string) {
    return this._get(scope, { exact: false });
  }

  /** Map over schema and transform leaf node by given handler */
  map<V>(handler: (node: SchemaNode) => V): TransformSchema<T, V> {
    return transformAccessControl(
      structuredClone(this._schema),
      this._options.prefix,
      this._options.spacer,
      (desc, scope) => {
        return handler(this.createSchemaNode(scope, desc));
      }
    ) as TransformSchema<T, V>;
  }

  /**
   * Create permit from schema with given granted scopes
   */
  permit(scopes: string[] | SchemaNode[]) {
    if (typeof scopes[0] === 'string' || scopes.length === 0) {
      return permit<T>(this, scopes as string[]);
    } else {
      const values = (scopes as SchemaNode[]).map((s) => s.scope);
      return permit<T>(this, values);
    }
  }
}
