import { describe, expect, test } from 'vitest';
import { createPermit, createSchema } from '..';
import schemaObjectDeep from './schema-deep';
import schemaObject from './schema-shallow';

describe('Permit Check Utility Test', () => {
  const schema = createSchema(schemaObject);
  const permit = schema.permit([
    'todo.view',
    'todo.action.toggle',
    'todo.action.edit',
    'todo.action.create',
  ]);

  test('expect exact scope "todo.view" check to be true', () => {
    expect(permit.check('todo.view').exact).toBe(true);
  });

  test('expect child of "todo.action" has any granted', () => {
    expect(permit.check('todo.action').hasAnyChildGranted).toBe(true);
  });

  test('expect child of "todo.action" has all granted to be false', () => {
    expect(permit.check('todo.action').hasAllChildGranted).toBe(false);
  });

  test('expect child of "todo.action" has all denied to be false', () => {
    expect(permit.check('todo.action').hasAllChildDenied).toBe(false);
  });

  test('expect child of "extra" has all denied to be false', () => {
    expect(permit.check('extra').hasAllChildDenied).toBe(false);
  });

  test('expect child of "extra" has all granted to be truth after changes', () => {
    permit.nodes.extra.sayHi.toggle();
    expect(permit.check('extra').hasAllChildGranted).toBe(true);
  });
});

describe('Permit Deep Use-case Test', () => {
  const schema = createSchema(schemaObjectDeep);
  const permit = schema.permit([
    'long.action.action.action.action.action.action.action.action.create',
  ]);

  test('expect long child to success checking value', () => {
    expect(
      permit.can.long.action.action.action.action.action.action.action.action
        .create
    ).toBeTruthy();
  });

  test('extract extra schema to be detached schema', () => {
    const child = schema.nodes.extra;
    const extraSchema = createSchema(child, { prefix: 'extra' });
    const extraPermit = extraSchema.permit([]);
    expect(extraPermit.export.all.length).toBe(1);
    expect(extraPermit.nodes.sayHi.scope).toBe('extra.sayHi');
    expect(extraPermit.can.sayHi).toBe(false);
  });

  test('createPermit shortcut also works with schema nodes getter', () => {
    const _p = createPermit(schema.nodes.extra, {
      prefix: 'extra',
      granted: ['extra.sayHi'],
    });
    expect(_p.export.all.length).toBe(1);
    expect(_p.nodes.sayHi.scope).toBe('extra.sayHi');
    expect(_p.can.sayHi).toBe(true);
  });
});

describe('Permit Simple Use-Case Test', () => {
  const schema = createSchema(schemaObject);
  const permit = schema.permit([
    'todo.view',
    'todo.action.toggle',
    'todo.action.edit',
    'todo.action.create',
    'extra.sayHi',
    'long.action.action.action.action.action.action.action.action.create',
  ]);

  /* ------------------------------ Type Checking ----------------------------- */
  // type Scopes = UnionScopes<typeof schemaObject>;
  // function selectScope(scope: Scopes) {
  //   console.log(scope)
  // }

  test('has total of 5 access granted', () => {
    expect(permit.granted.length).toBe(5);
  });

  test('has been granted access to todo.action.create', () => {
    expect(permit.can.todo.action.create).toBeTruthy();
  });

  test('denied access to delete todo', () => {
    expect(permit.can.todo.action.delete).toBeFalsy();
  });

  test('able to change granted access of todo.action.edit to be denied', () => {
    permit.nodes.todo.action.edit.toggle();
    expect(permit.can.todo.action.edit).toBeFalsy();
  });

  test('set todo.view to be denied', () => {
    permit.nodes.todo.view.granted = false;
    expect(permit.can.todo.view).toBeFalsy();
  });

  test('access that is not granted should be 3', () => {
    expect(permit.granted.length).toBe(3);
  });
});

describe('Permit Utility Test', () => {
  const permit = createPermit(schemaObject, { granted: ['todo.view'] });

  test('get utility test', () => {
    expect(permit.get).toBeDefined();

    const result = permit.get('todo.view');
    expect(result?.scope).toBe('todo.view');
    expect(result?.granted).toBeTruthy();
    expect(result?.desc).toBe('Enable user to view todo list page');
    expect(result?.toggle).toBeDefined();
  });

  test('has utility test', () => {
    expect(permit.has).toBeDefined();
    expect(permit.has('todo.view')).toBeTruthy();
  });

  test('search utility test', () => {
    expect(permit.search).toBeDefined();

    const found = [];
    permit.search('todo', (node) => found.push(node.scope));
    expect(found.length).toBe(5);
  });

  test('getChild utility test', () => {
    expect(permit.getChild).toBeDefined();

    const child = permit.getChild('todo.action');
    expect(child.length).toBe(4);
    expect(
      child.map((n) => n.scope).includes('todo.action.toggle')
    ).toBeTruthy();
  });

  test('reset utility test', () => {
    expect(permit.reset).toBeDefined();

    permit.nodes.todo.action.create.granted = true;
    expect(permit.can.todo.action.create).toBeTruthy();

    permit.reset();
    expect(permit.can.todo.action.create).toBeFalsy();
  });

  test('grant utility test', () => {
    expect(permit.grant).toBeDefined();

    permit.grant(
      'todo.action.create',
      'todo.action.edit',
      'todo.action.remove'
    );
    expect(permit.granted.length).toBe(3);
  });

  test('deny utility test', () => {
    expect(permit.deny).toBeDefined();

    permit.deny('todo.action.edit');
    expect(permit.granted.length).toBe(2);
  });

  test('export utility test', () => {
    expect(permit.export).toBeDefined();
    expect(permit.export.all.length).toBe(6);
    expect(permit.export.granted.length).toBe(2);
    expect(permit.export.denied.length).toBe(4);
  });

  test('granted utility test', () => {
    expect(permit.granted).toBeDefined();
    expect(permit.granted.length).toBe(2);
  });

  test('denied utility test', () => {
    expect(permit.denied).toBeDefined();
    expect(permit.denied.length).toBe(4);
  });

  test('nodes utility test', () => {
    expect(permit.nodes).toBeDefined();
    expect(permit.nodes.extra.sayHi).toBeDefined();
    expect(permit.nodes.extra.sayHi.scope).toBe('extra.sayHi');
    expect(permit.nodes.extra.sayHi.desc).toBe('Enable user to say hi');
    expect(permit.nodes.extra.sayHi.granted).toBeFalsy();
    expect(permit.nodes.extra.sayHi.toggle).toBeDefined();

    permit.nodes.extra.sayHi.toggle();
    expect(permit.nodes.extra.sayHi.granted).toBeTruthy();
  });

  test('can utility test', () => {
    expect(permit.can.extra.sayHi).toBeTruthy();
  });
});
